from dataclasses import dataclass
from typing import Any, Literal, Protocol

from .definitions import OPERATOR_PRECEDENCE
from .errors import (
    CommaOrClosingParenthesisExpectedError,
    MissingClosingParenthesisError,
    OpeningParenthesisExpectedError,
    UnexpectedEndOfExpressionError,
    UnexpectedNullFunctionNameError,
    UnexpectedNullIdentifierError,
    UnexpectedNullListError,
    UnexpectedNullLiteralError,
    UnexpectedNullNodeTypeError,
    UnexpectedNullOperandError,
    UnexpectedNullOperatorError,
    UnexpectedTokenError,
    UnknownNodeTypeError,
)
from .filter_tokenizer import (
    ODataFilterTokenizer,
    ODataFilterTokenizerProtocol,
    Token,
    TokenType,
)


@dataclass
class FilterNode:
    type_: Literal[
        'literal',
        'identifier',
        'operator',
        'function',
        'list',
        'value',
    ]
    value: Any | None = None
    left: 'FilterNode | None' = None
    right: 'FilterNode | None' = None
    arguments: list['FilterNode'] | None = None


class ODataFilterParserProtocol(Protocol):
    def parse(self, expr: str) -> FilterNode: ...

    def evaluate(self, node: FilterNode) -> str: ...


class ODataFilterParser:
    """Parser for OData V4 filter expressions."""

    __tokenizer: ODataFilterTokenizerProtocol
    """Tokenizer for the filter expression."""

    parse_null_identifier: bool
    """Whether to parse the null identifier."""

    def __init__(
        self,
        tokenizer: ODataFilterTokenizerProtocol | None = None,
        parse_null_identifier: bool = True,
    ) -> None:
        """Parser for OData V4 filter expressions.

        Parameters
        ----------
        tokenizer : ODataFilterTokenizerProtocol | None, optional
            Tokenizer, by default, it uses an instance of
            ``ODataFilterTokenizer``.
        parse_null_identifier : bool, optional
            Whether to parse the null identifier, by default True.

        """
        self.__tokenizer = tokenizer or ODataFilterTokenizer()
        self.parse_null_identifier = parse_null_identifier

    def set_tokenizer(self, tokenizer: ODataFilterTokenizerProtocol) -> None:
        """Set the tokenizer.

        Parameters
        ----------
        tokenizer : ODataFilterTokenizerProtocol
            Tokenizer.

        """
        self.__tokenizer = tokenizer  # pragma: no cover

    def parse(self, expr: str) -> FilterNode:
        """Parse a filter expression to an AST.

        Parameters
        ----------
        expr : str
            Filter expression to be parsed.

        Returns
        -------
        FilterNode
            AST representing the parsed filter expression.

        Examples
        --------
        >>> from odata_v4_query import ODataFilterParser
        >>> parser = ODataFilterParser()
        >>> ast = parser.parse("name eq 'John' and age gt 25")
        >>> ast
        FilterNode(type_='operator', value='and', left=FilterNode(...), right=FilterNode(...))

        """
        tokens = self.__tokenizer.tokenize(expr)
        if not tokens:
            return FilterNode(type_='value')

        return self._parse_expression(tokens)

    def evaluate(self, node: FilterNode) -> str:
        """Evaluate an AST to the original filter expression.

        Parameters
        ----------
        node : FilterNode
            AST representing the parsed filter expression.

        Returns
        -------
        str
            Filter expression.

        Raises
        ------
        UnexpectedNullNodeTypeError
            If node type is None.
        UnknownNodeTypeError
            If node type is unknown.

        Examples
        --------
        >>> from odata_v4_query import ODataFilterParser
        >>> parser = ODataFilterParser()
        >>> ast = parser.parse("name eq 'John' and age gt 25")
        >>> parser.evaluate(ast)
        "name eq 'John' and age gt 25"

        """
        if not node.type_:
            raise UnexpectedNullNodeTypeError(repr(node))

        handlers = {
            'literal': self._evaluate_literal,
            'identifier': self._evaluate_identifier,
            'list': self._evaluate_list,
            'operator': self._evaluate_operator,
            'function': self._evaluate_function,
            'value': lambda n: repr(n.value),
        }

        handler = handlers.get(node.type_)
        if not handler:
            raise UnknownNodeTypeError(node.type_)

        return handler(node)

    def _parse_expression(
        self,
        tokens: list[Token],
        precedence: int = 0,
    ) -> FilterNode:
        """Parse an expression using precedence climbing.

        This method implements the precedence climbing algorithm
        to parse expressions with operators of different precedence
        levels. It handles binary operators and builds an Abstract
        Syntax Tree (AST) representing the expression structure.

        Parameters
        ----------
        tokens : list[Token]
            List of tokens extracted from the filter expression.
        precedence : int, optional
            Minimum operator precedence level to consider,
            by default 0.

        Returns
        -------
        FilterNode
            AST node representing the parsed expression.

        Notes
        -----
        The precedence climbing algorithm works by:
        1. Parsing the leftmost expression first
        2. Looking ahead at the next operator
        3. If the operator has higher precedence than current level,
            recursively parse the right side
        4. Otherwise return the current expression

        """
        left = self._parse_primary(tokens)

        while tokens:
            token = tokens[0]
            if not isinstance(token, Token) or token.type_ != TokenType.OPERATOR:
                break

            op = token.value
            op_precedence = self._get_operator_precedence(op)  # type: ignore

            if op_precedence < precedence:
                break

            tokens.pop(0)  # consume operator
            right = self._parse_expression(tokens, op_precedence + 1)
            left = FilterNode(type_='operator', value=op, left=left, right=right)

        return left

    def _parse_primary(self, tokens: list[Token]) -> FilterNode:
        """Parse a primary expression.

        A primary expression is a literal, identifier, function call,
        or parenthesized expression. It is the smallest unit of an
        expression that can be evaluated.

        Parameters
        ----------
        tokens : list[Token]
            List of tokens extracted from the filter expression.

        Returns
        -------
        FilterNode
            AST representing the parsed primary expression.

        Raises
        ------
        UnexpectedEndOfExpressionError
            If an unexpected end of expression is reached.
        OpeningParenthesisExpectedError
            If an opening parenthesis is expected, but not found.
        UnexpectedTokenError
            If an unexpected token is encountered.

        """
        if not tokens:
            raise UnexpectedEndOfExpressionError

        token = tokens.pop(0)

        if token.type_ == TokenType.LITERAL:
            return FilterNode(type_='literal', value=token.value)

        if token.type_ == TokenType.IDENTIFIER:
            if token.value == 'null' and self.parse_null_identifier:
                return FilterNode(type_='identifier', value=None)

            return FilterNode(type_='identifier', value=token.value)

        if token.type_ == TokenType.FUNCTION:
            func_name = token.value

            if not tokens or tokens[0].type_ != TokenType.LPAREN:
                raise OpeningParenthesisExpectedError(str(func_name))

            tokens.pop(0)  # consume '('

            args = self._parse_list_values(tokens)
            return FilterNode(type_='function', value=func_name, arguments=args)

        if token.type_ == TokenType.LPAREN:
            values = self._parse_list_values(tokens)
            return FilterNode(type_='list', arguments=values)

        if token.type_ == TokenType.OPERATOR and token.value == 'not':
            expr = self._parse_expression(tokens, self._get_operator_precedence('not'))
            return FilterNode(type_='operator', value='not', right=expr)

        raise UnexpectedTokenError(str(token.value), token.position)

    def _parse_list_values(self, tokens: list[Token]) -> list[FilterNode]:
        """Parse a list of values without consuming the opening parenthesis.

        This method is used to parse the arguments of a function call
        and the values of the ``in`` operator.

        Parameters
        ----------
        tokens : list[Token]
            List of tokens extracted from the filter expression.

        Returns
        -------
        list[FilterNode]
            List of AST nodes representing the parsed values.

        Raises
        ------
        MissingClosingParenthesisError
            If closing parenthesis is missing.
        CommaOrClosingParenthesisExpectedError
            If no comma nor closing parenthesis is found.
        CommaOrClosingParenthesisExpectedError
            If token is neither a comma nor an closing parenthesis.

        """
        if not tokens:
            raise MissingClosingParenthesisError

        values = []

        # empty list
        if tokens[0].type_ == TokenType.RPAREN:
            tokens.pop(0)  # consume ')'
            return values

        while tokens:
            vnode = self._parse_expression(tokens)
            values.append(vnode)

            if not tokens:
                raise CommaOrClosingParenthesisExpectedError(
                    value=str(vnode.value),
                    expected_after=True,
                )

            # check for comma or closing parenthesis
            if tokens[0].type_ == TokenType.COMMA:
                tokens.pop(0)  # consume ','
            elif tokens[0].type_ == TokenType.RPAREN:
                tokens.pop(0)  # consume ')'
                break
            else:
                raise CommaOrClosingParenthesisExpectedError(value=str(tokens[0].value))

        return values

    def _get_operator_precedence(self, operator: str) -> int:
        """Return the precedence level of an operator.

        Parameters
        ----------
        operator : str
            Operator.

        Returns
        -------
        int
            Precedence level.

        """
        return OPERATOR_PRECEDENCE.get(operator, 0)

    def _evaluate_literal(self, node: FilterNode) -> str:
        """Evaluate a literal node.

        Parameters
        ----------
        node : FilterNode
            AST representing the parsed filter expression.

        Returns
        -------
        str
            Literal value.

        Raises
        ------
        UnexpectedNullLiteralError
            If node value is None.

        """
        if node.value is None:
            raise UnexpectedNullLiteralError(repr(node))

        # if it's numeric, don't add quotes
        if isinstance(node.value, (int, float)):
            return str(node.value)

        if isinstance(node.value, str):
            try:
                return str(float(node.value) if '.' in node.value else int(node.value))
            except ValueError:
                pass

        return f'{node.value!r}'

    def _evaluate_identifier(self, node: FilterNode) -> str:
        """Evaluate an identifier node.

        Parameters
        ----------
        node : FilterNode
            AST representing the parsed filter expression.

        Returns
        -------
        str
            Identifier value.

        Raises
        ------
        UnexpectedNullIdentifierError
            If node value is None.

        """
        if node.value is None:
            if self.parse_null_identifier:
                return 'null'

            raise UnexpectedNullIdentifierError(repr(node))

        return node.value

    def _evaluate_list(self, node: FilterNode, wrap_in_parentheses: bool = True) -> str:
        """Evaluate a list node.

        Parameters
        ----------
        node : FilterNode
            AST representing the parsed filter expression.
        wrap_in_parentheses : bool, optional
            Whether to wrap the list in parentheses, by default True.

        Returns
        -------
        str
            List value.

        Raises
        ------
        UnexpectedNullListError
            If node arguments is None.

        """
        if node.arguments is None:
            raise UnexpectedNullListError(repr(node))

        values = [self.evaluate(arg) for arg in node.arguments]

        if wrap_in_parentheses:
            return f'({", ".join(values)})'

        return ', '.join(values)

    def _evaluate_operator(self, node: FilterNode) -> str:
        """Evaluate an operator node.

        Parameters
        ----------
        node : FilterNode
            AST representing the parsed filter expression.

        Returns
        -------
        str
            Operator value.

        Raises
        ------
        UnexpectedNullOperatorError
            If node value is None.
        UnexpectedNullOperandError
            If node left or right is None.
        UnexpectedNullOperandError
            If node value is not ``not`` and node left
            or right is None.

        """
        if not node.value:
            raise UnexpectedNullOperatorError(repr(node))

        if node.value == 'not':
            if not node.right:
                raise UnexpectedNullOperandError(node.value)

            return f'not {self.evaluate(node.right)}'

        if not node.left or not node.right:
            raise UnexpectedNullOperandError(node.value)

        return f'{self.evaluate(node.left)} {node.value} {self.evaluate(node.right)}'

    def _evaluate_function(self, node: FilterNode) -> str:
        """Evaluate a function node.

        Parameters
        ----------
        node : FilterNode
            AST representing the parsed filter expression.

        Returns
        -------
        str
            Function value.

        Raises
        ------
        UnexpectedNullFunctionNameError
            If node value is None.

        """
        if not node.value:
            raise UnexpectedNullFunctionNameError(repr(node))

        args = self._evaluate_list(node, wrap_in_parentheses=False)
        return f'{node.value}({args})'



"""Definitions for OData V4 query options."""

DEFAULT_LIMIT = 100

DEFAULT_ORDERBY_DIRECTION = 'asc'

DEFAULT_FORMAT_OPTIONS = ('json', 'xml', 'csv', 'tsv')

# comparison
EQ = 'eq'
NE = 'ne'
GT = 'gt'
GE = 'ge'
LT = 'lt'
LE = 'le'
IN = 'in'
NIN = 'nin'

# logical
AND = 'and'
OR = 'or'
NOT = 'not'
NOR = 'nor'

# collection
HAS = 'has'

# functions
STARTSWITH = 'startswith'
ENDSWITH = 'endswith'
CONTAINS = 'contains'
SUBSTRING = 'substring'
TOLOWER = 'tolower'
TOUPPER = 'toupper'

COMPARISON_OPERATORS = (EQ, NE, GT, GE, LT, LE, IN, NIN)

LOGICAL_OPERATORS = (AND, OR, NOT, NOR)

FUNCTION_ARITY = {
    STARTSWITH: 2,
    ENDSWITH: 2,
    CONTAINS: 2,
    SUBSTRING: 3,
    TOLOWER: 1,
    TOUPPER: 1,
}

OPERATOR_ARITY = {
    # comparison
    EQ: 2,
    NE: 2,
    GT: 2,
    GE: 2,
    LT: 2,
    LE: 2,
    IN: 2,
    NIN: 2,
    # logical
    AND: 2,
    OR: 2,
    NOT: 1,
    NOR: 1,
    # collection
    HAS: 2,
}

OPERATOR_PRECEDENCE = {
    # comparison
    EQ: 4,
    NE: 4,
    GT: 4,
    GE: 4,
    LT: 4,
    LE: 4,
    IN: 4,
    NIN: 4,
    # collection
    HAS: 4,
    # logical
    NOT: 3,
    NOR: 3,
    AND: 2,
    OR: 1,
}



import pytest

from odata_v4_query.errors import (
    InvalidOrderDirectionError,
    NoNumericValueError,
    NoPositiveError,
    UnsupportedFormatError,
)
from odata_v4_query.query_parser import ODataQueryParser
from odata_v4_query.utils._func import remove_pagination_options


class TestODataQueryParser:
    parser = ODataQueryParser()

    def test_parse_url(self):
        url = 'https://example.com/odata?$count=true&$top=10&$skip=20'
        options = self.parser.parse_url(url)
        assert options.count
        assert options.top == 10
        assert options.skip == 20

    def test_parse_query_string(self):
        query_string = '$count=true&$top=10&$skip=20'
        options = self.parser.parse_query_string(query_string)
        assert options.count
        assert options.top == 10
        assert options.skip == 20

    def test_parse_query_params(self):
        query_params = {
            '$count': ['true'],
            '$top': ['10'],
            '$skip': ['20'],
            '$format': None,
        }
        options = self.parser.parse_query_params(query_params)
        assert options.count
        assert options.top == 10
        assert options.skip == 20
        assert options.format_ is None

    def test_parse_count(self):
        options = self.parser.parse_query_string('$count=true')
        assert options.count is True

        options = self.parser.parse_query_string('$count=false')
        assert options.count is False

    def test_shallow_clone_options(self):
        options = self.parser.parse_query_string(
            "$filter=name eq 'John'&$orderby=age desc"
        )
        cloned = options.clone(deep=False)
        assert cloned.filter_ is options.filter_
        assert cloned.orderby is options.orderby

    def test_deep_clone_options(self):
        options = self.parser.parse_query_string(
            "$filter=name eq 'John'&$orderby=age desc"
        )
        cloned_deep = options.clone(deep=True)
        assert cloned_deep.filter_ is not options.filter_
        assert cloned_deep.orderby is not options.orderby
        assert cloned_deep.filter_.value == options.filter_.value  # type: ignore
        assert cloned_deep.orderby[0].field == options.orderby[0].field  # type: ignore

    def test_parse_expand(self):
        options = self.parser.parse_query_string('$expand=field1,field2')
        assert options.expand == ['field1', 'field2']

    def test_parse_filter_and_evaluate(self):
        query_string = "$filter=name eq 'John' and age gt 25"
        options = self.parser.parse_query_string(query_string)
        EXPECTED = "name eq 'John' and age gt 25"

        assert options.filter_ is not None
        assert self.parser.evaluate(options) == EXPECTED
        assert self.parser.evaluate(options.filter_) == EXPECTED

        options.filter_ = None
        assert self.parser.evaluate(options) == ''

    def test_parse_format(self):
        options = self.parser.parse_query_string('$format=json')
        assert options.format_ == 'json'

    def test_parse_orderby(self):
        options = self.parser.parse_query_string('$orderby=name asc,age desc,,email')
        assert options.orderby is not None
        assert options.orderby[0].field == 'name'
        assert options.orderby[0].direction == 'asc'
        assert options.orderby[1].field == 'age'
        assert options.orderby[1].direction == 'desc'
        assert options.orderby[2].field == 'email'
        assert options.orderby[2].direction == 'asc'

    def test_parse_search(self):
        options = self.parser.parse_query_string('$search=John')
        assert options.search == 'John'

    def test_parse_select(self):
        options = self.parser.parse_query_string('$select=name,email')
        assert options.select == ['name', 'email']

    def test_parse_skip(self):
        options = self.parser.parse_query_string('$skip=10')
        assert options.skip == 10

    def test_parse_top(self):
        options = self.parser.parse_query_string('$top=10')
        assert options.top == 10

    def test_parse_page(self):
        options = self.parser.parse_query_string('$page=10')
        assert options.page == 10

    def test_no_numeric_skip(self):
        with pytest.raises(NoNumericValueError):
            self.parser.parse_query_string('$skip=ten')

    def test_no_numeric_top(self):
        with pytest.raises(NoNumericValueError):
            self.parser.parse_query_string('$top=ten')

    def test_no_numeric_page(self):
        with pytest.raises(NoNumericValueError):
            self.parser.parse_query_string('$page=ten')

    def test_unsupported_format(self):
        with pytest.raises(UnsupportedFormatError):
            self.parser.parse_query_string('$format=html')

    def test_invalid_order_direction(self):
        with pytest.raises(InvalidOrderDirectionError):
            self.parser.parse_query_string('$orderby=name xyz')

    def test_invalid_skip(self):
        with pytest.raises(NoPositiveError):
            self.parser.parse_query_string('$skip=-10')

    def test_invalid_top(self):
        with pytest.raises(NoPositiveError):
            self.parser.parse_query_string('$top=-10')

    def test_invalid_page(self):
        with pytest.raises(NoPositiveError):
            self.parser.parse_query_string('$page=-10')

    def test_remove_pagination_options(self):
        options = self.parser.parse_query_string('$top=10&$skip=5&$page=2')
        assert options.top == 10
        assert options.skip == 5
        assert options.page == 2

        remove_pagination_options(options)
        assert options.top is None
        assert options.skip is None
        assert options.page is None
